#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""These functions return information about the version of ISIS.

Most of the time, the only thing you'll probably need is the
version_info() function which returns an ISISversion tuple.
"""

# Copyright 2019-2020, Ross A. Beyer (rbeyer@seti.org)
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import collections
import datetime
import os
import re
from pathlib import Path

from kalasiris import environ


class ISISversion(collections.namedtuple('ISISversion',
                                         ['major', 'minor',
                                          'patch', 'releaselevel',
                                          'date'])):
    """This is a custom :func:`collections.namedtuple` which can
    contain ISIS version information.

    The first three elements, major, minor, and patch should be
    integers.  The fourth element, releaselevel, should be the
    string 'alpha', 'beta', or 'stable' or *None*.  And the fifth
    element, date, should be a :class:`datetime.date` object or *None*.
    That's what the functions in this module will return in an
    :class:`ISISversion` :func:`collections.namedtuple`.
    """


version_re = re.compile(r"(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+)")
date_re = re.compile(r"(?P<year>\d{4})-(?P<month>\d{1,2})-(?P<day>\d{1,2})")
date_yearlast_re = re.compile(r"(?P<month>\d{1,2})-(?P<day>\d{1,2})-(?P<year>\d{4})")  # noqa: E501
level_re = re.compile(r"^alpha|beta|stable")


def version_info() -> ISISversion:
    """Returned :func:`collections.namedtuple` of ISIS version information
    for the ISIS system underlying kalasiris.  If you want to answer,
    "What version of ISIS is being used?"  This is the function you're after.
    It is modeled after :func:`sys.version_info`."""
    return get_from_file(Path(environ['ISISROOT']) / 'version')


def get_from_string(s: str) -> ISISversion:
    """Read text and parse the contents for ISIS version information.

       This should parse ISIS version text as far back as ISIS 3.5.2.0,
       but possibly earlier.  It will return *None* values for releaselevel
       and date if it cannot parse them.  It will raise a :exc:`ValueError`
       if it cannot parse a version number.
    """

    # Version Matching
    match = version_re.search(s)
    if match:
        v = match.groupdict()
    else:
        raise ValueError(f"{s} did not match version regex: "
                         f"{version_re.pattern}")

    # Date Matching
    d = None
    d_match = date_re.search(s)
    if d_match:
        d = d_match.groupdict()
    else:
        d_match = date_yearlast_re.search(s)
        if d_match:
            d = d_match.groupdict()

    if d is not None:
        date = datetime.date(int(d['year']), int(d['month']), int(d['day']))
    else:
        date = None

    # Level Matching
    level = None
    l_match = level_re.search(s)
    if l_match:
        level = l_match.group()

    version = ISISversion(major=int(v['major']),
                          minor=int(v['minor']),
                          patch=int(v['patch']),
                          releaselevel=level,
                          date=date)
    return version


def get_from_file(file_path: os.PathLike) -> ISISversion:
    """Read an ISIS version file and parse the contents.

    This should parse version files as far back as ISIS 3.5.2.0,
    but possibly earlier.
    """

    v_text = Path(file_path).read_text()
    return get_from_string(v_text)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Utility functions for the test suite."""

# Copyright 2019-2020, Ross A. Beyer (rbeyer@seti.org)
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import collections
from pathlib import Path


def resource_check(*args):
    '''Checks to see if the files exist. And returns a tuple.

       Using the first element as the truth value and the second
       as the test value in a ``self.assertEqual(truth, test)``
       provides a much more useful failure message than a bunch
       of ``self.assertTrue(os.path.isfile(filename))``
    '''
    CheckReturn =  \
        collections.namedtuple('CheckReturn', ['truth', 'test'])

    truth = 'Missing Resources:'
    test = truth

    the_missing = []
    for path in map(Path, args):
        if not path.is_file():
            the_missing.append(path)
    if len(the_missing) > 0:
        test = '\n'.join(map(lambda x: '    ' + str(x), the_missing))
        test += "\n probably just need to 'make test-resources'"

    return CheckReturn(truth, test)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `kalasiris` package."""

# Copyright 2019-2020, Ross A. Beyer (rbeyer@seti.org)
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import contextlib
import os
import subprocess
import unittest
from unittest.mock import call, patch, MagicMock, Mock
from pathlib import Path

import kalasiris as isis
from .utils import resource_check as rc


# Hardcoding these, but I sure would like a better solution.
# IsisPreferences = os.path.join('test-resources', 'IsisPreferences')
HiRISE_img = Path('test-resources') / 'PSP_010502_2090_RED5_0.img'
run_real_files = True
run_real_files_reason = 'Tests on real files, and runs ISIS.'


@unittest.skipUnless(run_real_files, run_real_files_reason)
class TestResources(unittest.TestCase):
    '''Establishes that the test image exists.'''

    def test_resources(self):
        (truth, test) = rc(HiRISE_img)
        self.assertEqual(truth, test)


class Test_getkey_k(unittest.TestCase):

    def test_getkey_k(self):
        truth = 'HIRISE'
        gk = Mock(stdout=f'{truth}\n')
        with patch('kalasiris.k_funcs.isis.getkey', return_value=gk):
            key = isis.getkey_k('dummy.cub', 'Instrument', 'InstrumentId')
            self.assertEqual(truth, key)


@unittest.skipUnless(run_real_files, run_real_files_reason)
class Test_getkey_k_filesystem(unittest.TestCase):

    def setUp(self):
        self.cub = Path('test_getkey_k.cub')
        isis.hi2isis(HiRISE_img, to=self.cub)

    def tearDown(self):
        self.cub.unlink()
        with contextlib.suppress(FileNotFoundError):
            Path('print.prt').unlink()

    def test_getkey_k(self):
        truth = 'HIRISE'
        key = isis.getkey_k(self.cub, 'Instrument', 'InstrumentId')
        self.assertEqual(truth, key)


class Test_hi2isis_k(unittest.TestCase):

    @patch('kalasiris.k_funcs.isis.hi2isis')
    def test_with_to(self, m_hi2i):
        isis.hi2isis_k('dummy.img', to='dummy.cub')
        self.assertEqual(m_hi2i.call_args_list,
                         [call('dummy.img', to='dummy.cub')])

    @patch('kalasiris.k_funcs.isis.hi2isis')
    def test_without_to(self, m_hi2i):
        isis.hi2isis_k('dummy.img')
        self.assertEqual(m_hi2i.call_args_list, [call('dummy.img',
                                                      to=Path('dummy.cub'))])


@unittest.skipUnless(run_real_files, run_real_files_reason)
class Test_hi2isis_k_filesystem(unittest.TestCase):

    def setUp(self):
        self.img = HiRISE_img

    def tearDown(self):
        with contextlib.suppress(FileNotFoundError):
            Path('print.prt').unlink()

    def test_with_to(self):
        tocube = Path('test_hi2isis_k.cub')
        isis.hi2isis_k(self.img, to=tocube)
        self.assertTrue(tocube.is_file())
        tocube.unlink()

    def test_without_to(self):
        tocube = Path(self.img).with_suffix('.cub')
        isis.hi2isis_k(self.img)
        self.assertTrue(tocube.is_file())
        tocube.unlink()


class Test_hist_k(unittest.TestCase):

    @patch('kalasiris.k_funcs.isis.hist')
    def test_run(self, m_hist):
        hist_txt = 'This is hist output.'
        filelike = Mock()
        filelike.read = Mock(return_value=hist_txt)
        with patch('kalasiris.k_funcs.isis.tempfile.NamedTemporaryFile',
                   return_value=filelike):
            hist_as_string = isis.hist_k('dummy.cub')
            self.assertEqual(hist_as_string, hist_txt)

    def test_fail(self):
        # ISIS hist_k needs at least a FROM=, giving it nothing:
        self.assertRaises(subprocess.CalledProcessError, isis.hist_k)


@unittest.skipUnless(run_real_files, run_real_files_reason)
class Test_hist_k_filesystem(unittest.TestCase):

    def setUp(self):
        self.cube = Path('test_hist.cub')
        isis.hi2isis(HiRISE_img, to=self.cube)

    def tearDown(self):
        self.cube.unlink()
        with contextlib.suppress(FileNotFoundError):
            Path('print.prt').unlink()

    def test_run(self):
        hist_as_string = isis.hist_k(self.cube)
        self.assertTrue(hist_as_string.startswith('Cube'))

    def test_run_with_to(self):
        text_file = Path('test_hist.hist')
        hist_as_string = isis.hist_k(self.cube, to=text_file)
        self.assertTrue(text_file.is_file())
        self.assertTrue(hist_as_string.startswith('Cube'))
        text_file.unlink()


class Test_stats_k(unittest.TestCase):

    def test_stats_k(self):
        stats_text = '''Group = Results
  From                    = PSP_010502_2090_RED4_1.cub
  Band                    = 1
  Average                 = 6498.477293457
  StandardDeviation       = 181.94624138776
  Variance                = 33104.434755134
  Median                  = 6497.0
  Mode                    = 6534.0
  Skew                    = 0.024358185897602
  Minimum                 = 4117.0
  Maximum                 = 8207.0
  Sum                     = 13308881497.0
  TotalPixels             = 2048000
  ValidPixels             = 2048000
  OverValidMaximumPixels  = 0
  UnderValidMinimumPixels = 0
  NullPixels              = 0
  LisPixels               = 0
  LrsPixels               = 0
  HisPixels               = 0
  HrsPixels               = 0
End_Group
'''
        cp = Mock(stdout=stats_text)
        with patch('kalasiris.k_funcs.isis.stats', return_value=cp):
            d = isis.stats_k('foo')
            self.assertEqual(20, len(d))
            self.assertEqual(d['Average'], '6498.477293457')

    @unittest.skipUnless(run_real_files, run_real_files_reason)
    def test_stats_k_file(self):
        cub = Path('test_stats_k.cub')
        isis.hi2isis(HiRISE_img, to=cub)

        d = isis.stats_k(cub)
        self.assertEqual(d['TotalPixels'], '2048000')

        cub.unlink()
        with contextlib.suppress(FileNotFoundError):
            Path('print.prt').unlink()


class Test_cubeit_k(unittest.TestCase):

    @patch('kalasiris.k_funcs.isis.cubeit')
    def test_cubeit_k(self, m_cubeit):
        from_name = 'temp_fromlist.txt'
        m_context = Mock(__enter__=Mock(return_value=from_name),
                         __exit__=Mock())
        m_temp = MagicMock(return_value=m_context)
        with patch('kalasiris.k_funcs.isis.fromlist.temp', m_temp):
            isis.cubeit_k(['a.cub', 'b.cub', 'c.cub'], to='stacked.cub')
            m_temp.assert_called_with(['a.cub', 'b.cub', 'c.cub'])
            self.assertEqual(m_cubeit.call_args_list,
                             [call(fromlist=from_name, to='stacked.cub')])

    @unittest.skipUnless(run_real_files, run_real_files_reason)
    def test_cubeit_k_files(self):
        a_cube = 'test_cubeit_a.cub'
        isis.makecube(to=a_cube, value=1, samples=2, lines=2, bands=1)
        b_cube = 'test_cubeit_b.cub'
        isis.makecube(to=b_cube, value=1, samples=2, lines=2, bands=1)
        c_cube = 'test_cubeit_c.cub'
        isis.makecube(to=c_cube, value=1, samples=2, lines=2, bands=1)
        s_cube = 'test_cubeit_stacked.cub'
        isis.cubeit_k([a_cube, b_cube, c_cube], to=s_cube)
        for f in (a_cube, b_cube, c_cube, s_cube):
            os.unlink(f)
        with contextlib.suppress(FileNotFoundError):
            Path('print.prt').unlink()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `PathSet` class."""

# Copyright 2019-2020, Ross A. Beyer (rbeyer@seti.org)
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import contextlib
import unittest
from pathlib import Path

import kalasiris as isis


run_real_files = True
run_real_files_reason = 'Tests on real files.'


class TestPathSet(unittest.TestCase):

    def setUp(self):
        self.paths = (Path('test_PathSet_1.cub'),
                      Path('test_PathSet_2.txt'),
                      Path('test_PathSet_3.foo'))

    def tearDown(self):
        with contextlib.suppress(FileNotFoundError):
            for p in self.paths:
                p.unlink()

    def test_init(self):
        ps = isis.PathSet()
        self.assertIsInstance(ps, isis.PathSet)

    def test_init_iterable(self):
        ps = isis.PathSet(self.paths)
        self.assertEqual(3, len(ps))

    def test_init_fail(self):
        self.assertRaises(TypeError, isis.PathSet, 'a string')

    def test_add(self):
        ps = isis.PathSet()
        path1 = self.paths[0]
        added_path = ps.add(path1)
        self.assertEqual(path1, added_path)

    def test_add_Error(self):
        ps = isis.PathSet()
        p0 = self.paths[0]
        ps.add(p0)
        self.assertRaises(TypeError, ps.add, 'not a Path')
        self.assertRaises(ValueError, ps.add, p0)

    @unittest.skipUnless(run_real_files, run_real_files_reason)
    def test_unlink(self):
        ps = isis.PathSet(self.paths)
        for p in ps:
            p.touch()
        ps.unlink()
        for p in ps:
            self.assertFalse(p.exists())



# -*- coding: utf-8 -*-

"""Top-level package for kalasiris."""

__author__ = """Ross A. Beyer"""
__email__ = 'rbeyer@seti.org'
__version__ = '1.7.0'

from .kalasiris import *                      # noqa: F401,F403
from .k_funcs import *                        # noqa: F401,F403
from .Histogram import Histogram              # noqa: F401
from .PathSet import PathSet                  # noqa: F401
import kalasiris.cube                         # noqa: F401
import kalasiris.cubenormfile                 # noqa: F401
import kalasiris.fromlist                     # noqa: F401
import kalasiris.specialpixels                # noqa: F401



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for the `cube` module."""

# Copyright 2020, Ross A. Beyer (rbeyer@seti.org)
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import contextlib
import unittest
from pathlib import Path

import kalasiris as isis
from .utils import resource_check as rc

run_real_files = True
run_real_files_reason = 'Tests on real files, and runs ISIS.'

# Hardcoding this, but I sure would like a better solution.
HiRISE_img = Path('test-resources') / 'PSP_010502_2090_RED5_0.img'
img = HiRISE_img


@unittest.skipUnless(run_real_files, run_real_files_reason)
class TestResources(unittest.TestCase):
    '''Establishes that the test image exists.'''

    def test_resources(self):
        (truth, test) = rc(img)
        self.assertEqual(truth, test)


class TestBasic(unittest.TestCase):

    def test_get_start_size(self):
        d = {'StartByte': '10', 'Bytes': '20'}
        self.assertEqual((9, 20), isis.cube._get_start_size(d))


@unittest.skipUnless(run_real_files, run_real_files_reason)
class TestTable(unittest.TestCase):

    def setUp(self):
        self.cube = Path('test_TableRead.cub')
        isis.hi2isis(img, to=self.cube)

    def tearDown(self):
        with contextlib.suppress(FileNotFoundError):
            Path('print.prt').unlink()
        self.cube.unlink()

    def test_read_table_data(self):
        self.assertRaises(ValueError, isis.cube.read_table_data, self.cube)
        self.assertRaises(KeyError, isis.cube.read_table_data, self.cube,
                          table_name='Non-Existant Table Name')

        table_bytes = isis.cube.read_table_data(
            self.cube, table_name='HiRISE Calibration Ancillary')
        self.assertEqual(7440, len(table_bytes))

        label = {'StartByte': '4169490', 'Bytes': '126976'}
        table_bytes = isis.cube.read_table_data(
            self.cube, label=label)
        self.assertEqual(int(label['Bytes']), len(table_bytes))

    def test_parse_table(self):
        table_bytes = isis.cube.read_table_data(
            self.cube, table_name='HiRISE Calibration Ancillary')

        fields = [{'Name': 'GapFlag', 'Type': 'Integer', 'Size': '1'},
                  {'Name': 'LineNumber', 'Type': 'Integer', 'Size': '1'},
                  {'Name': 'BufferPixels', 'Type': 'Integer', 'Size': '12'},
                  {'Name': 'DarkPixels', 'Type': 'Integer', 'Size': '16'}]

        table = isis.cube.parse_table(table_bytes, fields)

        self.assertEqual(255, table['GapFlag'][0])
        self.assertEqual(9, table['LineNumber'][9])
        self.assertEqual(1359, table['BufferPixels'][0][0])
        self.assertEqual(1290, table['DarkPixels'][0][9])

    def test_get_table(self):
        try:
            import pvl  # noqa F401

            table = isis.cube.get_table(self.cube,
                                        'HiRISE Calibration Ancillary')
            self.assertEqual(255, table['GapFlag'][0])
            self.assertEqual(9, table['LineNumber'][9])
            self.assertEqual(1359, table['BufferPixels'][0][0])
            self.assertEqual(1290, table['DarkPixels'][0][9])

        except ImportError:
            self.assertRaises(ImportError, isis.cube.get_table, self.cube,
                              'HiRISE Calibration Ancillary')

    def test_overwrite_table_data(self):
        data = bytes(10)
        self.assertRaises(ValueError,
                          isis.cube.overwrite_table_data, self.cube, data)
        self.assertRaises(KeyError,
                          isis.cube.overwrite_table_data, self.cube, data,
                          table_name='Non-Existant Table Name')
        self.assertRaises(ValueError,
                          isis.cube.overwrite_table_data, self.cube, data,
                          table_name='HiRISE Calibration Ancillary')

        # Write zeros into the table:
        table_data = bytes(7440)
        isis.cube.overwrite_table_data(
            self.cube, table_data, table_name='HiRISE Calibration Ancillary')

        # Read the table back, verify a zero:
        table_bytes = isis.cube.read_table_data(
            self.cube, table_name='HiRISE Calibration Ancillary')
        fields = [{'Name': 'A', 'Type': 'Integer', 'Size': '1'},
                  {'Name': 'B', 'Type': 'Integer', 'Size': '1'},
                  {'Name': 'C', 'Type': 'Integer', 'Size': '12'},
                  {'Name': 'D', 'Type': 'Integer', 'Size': '16'}]
        table = isis.cube.parse_table(table_bytes, fields)
        self.assertEqual(0, table['C'][0][0])

    def test_encode_table(self):
        table = {'Foo': [1, 2, 3, 4],
                 'Bar': [1.1, 2.2, 3.3, 4.4],
                 'Boo': ['a', 'b', 'c', 'd']}
        fields = [{'Name': 'Foo', 'Type': 'Integer', 'Size': '1'},
                  {'Name': 'Bar', 'Type': 'Real', 'Size': '1'},
                  {'Name': 'Boo', 'Type': 'Text', 'Size': '1'}]

        bad_table = dict(table)
        bad_table['TooLong'] = ['a', 'b', 'c', 'd', 'e']
        self.assertRaises(IndexError, isis.cube.encode_table,
                          bad_table, fields)

        bad_fields = list(fields)
        bad_fields.append({'Name': 'TooLong', 'Type': 'Text', 'Size': '1'})
        self.assertRaises(KeyError, isis.cube.encode_table,
                          table, bad_fields)

        data = isis.cube.encode_table(table, fields)
        # print(len(data))
        # print(data)

        parsed_tab = isis.cube.parse_table(data, fields)
        # print(tab)
        self.assertEqual(table['Foo'], parsed_tab['Foo'])

    def test_overwrite_table(self):
        table_bytes = isis.cube.read_table_data(
            self.cube, table_name='HiRISE Calibration Ancillary')

        fields = [{'Name': 'GapFlag', 'Type': 'Integer', 'Size': '1'},
                  {'Name': 'LineNumber', 'Type': 'Integer', 'Size': '1'},
                  {'Name': 'BufferPixels', 'Type': 'Integer', 'Size': '12'},
                  {'Name': 'DarkPixels', 'Type': 'Integer', 'Size': '16'}]

        table = isis.cube.parse_table(table_bytes, fields)

        for i, ln in enumerate(table['LineNumber']):
            table['LineNumber'][i] = 27

        try:
            import pvl  # noqa F401

            isis.cube.overwrite_table(self.cube,
                                      'HiRISE Calibration Ancillary',
                                      table)

            p_tab = isis.cube.get_table(self.cube,
                                        'HiRISE Calibration Ancillary')

            self.assertEqual(255, p_tab['GapFlag'][0])
            self.assertEqual(27, p_tab['LineNumber'][9])
            self.assertEqual(1359, p_tab['BufferPixels'][0][0])
            self.assertEqual(1290, p_tab['DarkPixels'][0][9])

        except ImportError:
            self.assertRaises(ImportError, isis.cube.overwrite_table,
                              self.cube, 'HiRISE Calibration Ancillary',
                              table)



# -*- coding: utf-8 -*-

"""Constants for Isis Special Pixels.

The constants are provided as SpecialPixel :func:collection.namedtuple:
objects, and have the following keys: Min, Null, Lrs, Lis, His,
Hrs, and Max.

Null
    A null pixel indicates no data was collected at the particular
    location.

Lis
    A low instrument saturation pixel occurred meaning the
    instrument readout was at its lowest possible value.

His
    A high instrument saturation pixel occurred meaning the
    instrument readout was at its highest possible value.

Lrs
    A low representation saturation pixel occurred meaning a
    program computed a new pixel value at this location that
    was lower than native bit-type for the file (e.g., less
    than zero for an 8-bit file).

Hrs:
    A high representation saturation pixel occurred meaning a
    program computed a new pixel value at this location that
    was greater than native bit-type for the file (e.g., greater
    than 255 for an 8-bit file).

The above definitions come from the `Logical Cube Format Guide
<https://isis.astrogeology.usgs.gov/documents/LogicalCubeFormatGuide/LogicalCubeFormatGuide.html>`_
and the numerical values come from the ISIS `SpecialPixel.h
<https://github.com/USGS-Astrogeology/ISIS3/blob/dev/isis/src/base/objs/SpecialPixel/SpecialPixel.h>`_ file.

In order to be able to know whether value should result in a
*Lrs* or *Hrs* value, we must also know:

Min
    The minimum valid value for a pixel.

Max
    The maximum valid value for a pixel.
"""  # noqa

# Copyright 2015, William Trevor Olson
# Copyright 2020, Ross A. Beyer (rbeyer@seti.org)
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import collections
import struct
import sys


SpecialPixels = collections.namedtuple('SpecialPixels',
                                       ['Min', 'Null',
                                        'Lrs', 'Lis',
                                        'His', 'Hrs',
                                        'Max'])

# 1-byte special pixel values from SpecialPixel.h
UnsignedByte = SpecialPixels(Min=1, Null=0, Lrs=0, Lis=0,
                             His=255, Hrs=255, Max=254)

# 2-byte unsigned special pixel values from SpecialPixel.h
UnsignedWord = SpecialPixels(Min=3, Null=0, Lrs=1, Lis=2,
                             His=65534, Hrs=65535, Max=65522)

# 2-byte signed special pixel values from SpecialPixel.h
SignedWord = SpecialPixels(Min=-32752, Null=-32768,
                           Lrs=-32767, Lis=-32766,
                           His=-32765, Hrs=-32764, Max=32767)

# 4-byte unsigned special pixel values from SpecialPixel.h
# This deals with unsigned long int values.  Not sure why there
# is so much space between 'Max' and the 'His' and 'Hrs' values, but
# that's what's in SpecialPixel.h.
UnsignedInteger = SpecialPixels(Min=3, Null=0, Lrs=1, Lis=2,
                                His=4294967294, Hrs=4294967295,
                                Max=4294967282)

# This was in the original pysis, but I don't see it in SpecialPixel.h?
# The 'Max' value indicates that it is like a long int, but the 'Min'
# and other negative values are far from the possible representation
# minimum of a long int, so not sure what's going on here.
SignedInteger = SpecialPixels(Min=-8388614, Null=-8388613,
                              Lrs=-8388612, Lis=-8388611,
                              His=-8388610, Hrs=-8388609, Max=2147483647)

# 4-byte special pixel values for IEEE floating point from SpecialPixel.h
Real = SpecialPixels(
    Min=struct.unpack('>f', bytes.fromhex('FF7FFFFA'))[0],
    Null=struct.unpack('>f', bytes.fromhex('FF7FFFFB'))[0],
    Lrs=struct.unpack('>f', bytes.fromhex('FF7FFFFC'))[0],
    Lis=struct.unpack('>f', bytes.fromhex('FF7FFFFD'))[0],
    His=struct.unpack('>f', bytes.fromhex('FF7FFFFE'))[0],
    Hrs=struct.unpack('>f', bytes.fromhex('FF7FFFFF'))[0],
    Max=sys.float_info.max)

# 8-byte special pixel values for IEEE floating point from SpecialPixel.h
Double = SpecialPixels(
    Min=struct.unpack('>d', bytes.fromhex('FFEFFFFF FFFFFFFA'))[0],
    Null=struct.unpack('>d', bytes.fromhex('FFEFFFFF FFFFFFFB'))[0],
    Lrs=struct.unpack('>d', bytes.fromhex('FFEFFFFF FFFFFFFC'))[0],
    Lis=struct.unpack('>d', bytes.fromhex('FFEFFFFF FFFFFFFD'))[0],
    His=struct.unpack('>d', bytes.fromhex('FFEFFFFF FFFFFFFE'))[0],
    Hrs=struct.unpack('>d', bytes.fromhex('FFEFFFFF FFFFFFFF'))[0],
    Max=sys.float_info.max)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for the `fromlist` class."""

# Copyright 2019-2020, Ross A. Beyer (rbeyer@seti.org)
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import contextlib
import os
import unittest
from unittest.mock import Mock, patch
from pathlib import Path

import kalasiris as isis


run_real_files = True
run_real_files_reason = 'Tests on real files.'


class TestFromList(unittest.TestCase):

    def setUp(self):
        self.list = ['a.cub', 'b.cub', 'c.cub']

    @patch('kalasiris.fromlist.builtins.print')
    def test_print(self, m_print):
        isis.fromlist.print(self.list)
        self.assertEqual(m_print.call_args_list[0][0][0], self.list[0])
        self.assertEqual(m_print.call_args_list[1][0][0], self.list[1])
        self.assertEqual(m_print.call_args_list[2][0][0], self.list[2])

    @patch('kalasiris.fromlist.builtins.print')
    def test_print_fl(self, m_print):
        isis.fromlist.print_fl(self.list)
        self.assertEqual(m_print.call_args_list[0][0][0], self.list[0])
        self.assertEqual(m_print.call_args_list[1][0][0], self.list[1])
        self.assertEqual(m_print.call_args_list[2][0][0], self.list[2])

    @patch('kalasiris.fromlist.print_fl')
    def test_make(self, m_print_fl):
        m_filelike = Mock()
        m_filelike.name = 'dummy.txt'
        with patch('kalasiris.fromlist.tempfile.NamedTemporaryFile',
                   return_value=m_filelike):
            temp_file = isis.fromlist.make(self.list)
            self.assertEqual(temp_file, Path(m_filelike.name))
            m_print_fl.called_once_with(self.list, file=m_filelike)

    @patch('kalasiris.fromlist.print_fl')
    def test_make_wfile(self, m_print_fl):
        m_filelike = Mock()
        m_filelike.name = 'dummy.txt'
        with patch('kalasiris.fromlist.open', return_value=m_filelike):
            temp_file = isis.fromlist.make(self.list, m_filelike.name)
            self.assertEqual(temp_file, Path(m_filelike.name))
            m_print_fl.called_once_with(self.list, file=m_filelike)


@unittest.skipUnless(run_real_files, run_real_files_reason)
class TestFromList_filesystem(unittest.TestCase):

    def setUp(self):
        self.list = ['a.cub', 'b.cub', 'c.cub']
        self.text = 'a.cub\nb.cub\nc.cub\n'
        self.path = Path('test_fromlist.txt')

    def tearDown(self):
        with contextlib.suppress(FileNotFoundError):
            self.path.unlink()

    def test_print(self):
        with open(self.path, 'w') as f:
            isis.fromlist.print(self.list, f)
        self.assertTrue(self.path.exists())
        self.assertEqual(self.path.read_text(), self.text)

    def test_print_fl(self):
        with open(self.path, 'w') as f:
            isis.fromlist.print_fl(self.list, f)
        self.assertTrue(self.path.exists())
        self.assertEqual(self.path.read_text(), self.text)

    def test_make(self):
        temp_file = isis.fromlist.make(self.list)
        self.assertTrue(temp_file.exists())
        self.assertEqual(temp_file.read_text(), self.text)
        temp_file.unlink()

    def test_make_wfile(self):
        isis.fromlist.make(self.list, self.path)
        self.assertTrue(self.path.exists())
        self.assertEqual(self.path.read_text(), self.text)

    def test_open_fl(self):
        with isis.fromlist.open_fl(self.list) as f:
            self.assertEqual(f.read(), self.text)
            filename = f.name
        self.assertFalse(os.path.exists(filename))

    def test_open_fl_wfile(self):
        with isis.fromlist.open_fl(self.list, self.path) as f:
            self.assertEqual(f.read(), self.text)
        self.assertTrue(self.path.exists())

    def test_temp(self):
        with isis.fromlist.temp(self.list) as f:
            self.assertEqual(f.read_text(), self.text)
            filename = f
            self.assertTrue(filename.exists())
        self.assertFalse(filename.exists())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `kalasiris` package."""

# Copyright 2019-2020, Ross A. Beyer (rbeyer@seti.org)
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import importlib
import os
import unittest
import kalasiris.kalasiris


class TestHOME(unittest.TestCase):
    '''Not all platforms have a $HOME environment variable.'''

    def test_HOME(self):
        if 'HOME' in os.environ:
            del os.environ['HOME']

        self.assertTrue(importlib.reload(kalasiris.kalasiris))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2019-2020, Ross A. Beyer (rbeyer@seti.org)
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import collections
import csv
import subprocess
from .k_funcs import hist_k


class Histogram(collections.abc.Sequence):
    """Reads the output from ISIS hist and provides it as a sequence.

    The resulting Histogram object primarily behaves like a list, where
    that list represents the rows of the ISIS hist output.  Each of those
    rows is a :func:`collections.namedtuple` which contains the elements
    of each row, referenced by their column names.

    The Histogram object also has some dictionary-like capabilities, in
    order to get at the values listed in the ISIS hist output in the
    section before the numerical output.
    """

    def __init__(self, histinfo):
        self.histinfo = histinfo

        try:
            (self.dictionary, self.headers, self.hist_list) \
                = self.parse(histinfo)
        except StopIteration:
            try:
                (self.dictionary, self.headers, self.hist_list) \
                    = self.parse(hist_k(histinfo))
            except subprocess.CalledProcessError:
                with open(histinfo, 'r') as f:
                    (self.dictionary, self.headers, self.hist_list) \
                        = self.parse(f.read())

    def __str__(self):
        return(str(self.dictionary))

    def __repr__(self):
        return (f'{self.__class__.__name__}(\'{self.histfile}\')')

    def __len__(self):
        return len(self.hist_list)

    def __getitem__(self, key):
        try:
            return self.dictionary[key]
        except KeyError:
            return self.hist_list[key]

    def __iter__(self):
        return self.hist_list.__iter__()

    def __contains__(self, item):
        if item in self.dictionary:
            return True
        else:
            return(item in self.hist_list)

    def keys(self):
        """Gets the keys from the initial portion of the hist output file.

           These will be items like 'Cube', 'Band', 'Average', etc.
        """
        return self.dictionary.keys()

    def values(self):
        """Gets the values from the initial portion of the hist output file."""
        return self.dictionary.values()

    @staticmethod
    def parse(histinfo: str) -> tuple:
        """Takes a string (expecting the output of ISIS ``hist``), and
        parses the output.

        A three-element namedtuple is returned: the first element
        is a *dictionary* of the name:value information at the
        top of the file, the second element is a *list* of the
        of the fields that decorate the top of the histogram
        rows, and the third element is a *list* of HistRow
        :func:`collections.namedtuple` objects that represent each
        row of the hist output.

        The contents of the file that results from ISIS ``hist``
        look like this::

             Cube:           PSP_010502_2090_RED5_0.EDR_Stats.cub
             Band:           1
             Average:        6490.68
             [... other lines like this ...]
             His Pixels:      0
             Hrs Pixels:      0


             DN,Pixels,CumulativePixels,Percent,CumulativePercent
             3889,1,1,4.88281e-05,4.88281e-05
             3924,1,2,4.88281e-05,9.76563e-05
             3960,2,4,9.76563e-05,0.000195313
             [... more comma-separated lines like above ...]

        But this function sees it like this::

             k: v
             k: v
             k: v
             [... other lines like this ...]
             k: v
             k: v


             h,h,h,h,h
             n,n,n,n,n
             n,n,n,n,n
             n,n,n,n,n
             [... more comma-separated lines like above ...]

        Where each of the letters above is a string value that the parser
        reads.

        First, it takes all of the ``k`` and ``v`` elements and saves them
        as the keys and values in the returned dictionary.

        Second, the ``h`` elements are returned them as the list of
        fieldnames.

        Third, it reads the lines with ``n`` and stores them as
        ``namedtuples`` in the returned list.
        """
        d = dict()
        fieldnames = []
        hist_rows = []
        for line in filter(lambda x: ':' in x, str(histinfo).splitlines()):
            (k, v) = line.split(':')
            # d[k.strip()] = v.strip()
            d.setdefault(k.strip(), v.strip())

        reader = csv.reader(filter(lambda x: ',' in x,
                                   str(histinfo).splitlines()))
        fieldnames = next(reader)

        HistRow = collections.namedtuple('HistRow', fieldnames)
        for row in map(HistRow._make, reader):
            hist_rows.append(row)

        HistParsed = collections.namedtuple('HistParsed',
                                            ['info', 'fieldnames', 'data'])
        return(HistParsed(d, fieldnames, hist_rows))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `kalasiris` package."""

# Copyright 2019-2020, Ross A. Beyer (rbeyer@seti.org)
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import contextlib
import subprocess
import unittest
from pathlib import Path

from kalasiris.kalasiris import _get_isis_program_names as gipn
import kalasiris.kalasiris as isis
from .utils import resource_check as rc


run_real_files = True
run_real_files_reason = 'Tests on real files, and runs ISIS.'

# Hardcoding this, but I sure would like a better solution.
HiRISE_img = Path('test-resources') / 'PSP_010502_2090_RED5_0.img'
img = HiRISE_img


@unittest.skipUnless(run_real_files, run_real_files_reason)
class TestResources(unittest.TestCase):
    '''Establishes that the test image exists.'''

    def test_resources(self):
        (truth, test) = rc(img)
        self.assertEqual(truth, test)


class TestParams(unittest.TestCase):

    def test_format(self):
        t = ('isisprogram', 'from=foo.cub')
        p = {'to': 'to.cub', 'check': False, 'value': 3.0}
        cmd = list(t)
        truth = list(t)
        truth.extend(['to=to.cub', 'check=False', 'value=3.0'])

        cmd.extend(map(isis.param_fmt, p.keys(), p.values()))
        self.assertEqual(truth, cmd)

    @unittest.skip('Fires up the gui, not sure how to test properly.')
    def test_no_args(self):
        print('\n  about to getkey()')
        isis.getkey()
        # isis.getkey('gui__') does the same thing.
        print('  just got back from getkey()')

    @unittest.skip('Fires up the gui, not sure how to test properly.')
    def test_passthrough(self):
        to_cube = Path('test_passthrough.cub')
        isis.hi2isis(HiRISE_img, to=to_cube)
        print('\n  about to call qview()')
        isis.qview(to_cube)
        print('  just got back from qview()')
        to_cube.unlink()

    def test_reserved_param(self):
        t = 'isisprogram'
        p = {'help__': 'parameter'}
        cmd = [t]
        truth = [t]
        truth.append('-help=parameter')

        cmd.extend(map(isis.param_fmt, p.keys(), p.values()))
        self.assertEqual(truth, cmd)

    def test_reserved_nokey(self):
        cp = isis.getkey('help__').stdout.split()

        self.assertEqual('FROM', cp[0])


class Test_get_isis_program_names(unittest.TestCase):

    @unittest.skipUnless(run_real_files, run_real_files_reason)
    def test_get_names(self):
        # for n in gipn():
        #     print(n)
        # s = sum(1 for _ in gipn())
        # print(f'How many programs: {s}')
        self.assertIn('cam2map', gipn())


@unittest.skipUnless(run_real_files, run_real_files_reason)
class Test_hi2isis(unittest.TestCase):

    def setUp(self):
        self.img = img

    def tearDown(self):
        with contextlib.suppress(FileNotFoundError):
            Path('print.prt').unlink()

    def test_hi2isis(self):
        tocube = Path('test_hi2isis.cub')
        isis.hi2isis(self.img, to=tocube)
        self.assertTrue(tocube.is_file())
        tocube.unlink()


@unittest.skipUnless(run_real_files, run_real_files_reason)
class Test_getkey(unittest.TestCase):

    def setUp(self):
        self.cub = Path('test_getkey.cub')
        isis.hi2isis(img, to=self.cub)

    def tearDown(self):
        self.cub.unlink()
        with contextlib.suppress(FileNotFoundError):
            Path('print.prt').unlink()

    def test_getkey(self):
        truth = 'HIRISE'
        key = isis.getkey(self.cub, grpname='Instrument',
                          keyword='InstrumentId').stdout.strip()
        self.assertEqual(truth, key)

    def test_getkey_fail(self):
        # Pixels doesn't have InstrumentId, should fail
        self.assertRaises(subprocess.CalledProcessError,
                          isis.getkey, self.cub,
                          grpname='Pixels', keyword='InstrumentId')

    def test_getkey_k_fail(self):
        # Calling getkey with getkey_k syntax will fail
        self.assertRaises(IndexError,
                          isis.getkey, self.cub,
                          'Instrument', 'InstrumentId')


@unittest.skipUnless(run_real_files, run_real_files_reason)
class Test_histat(unittest.TestCase):

    def setUp(self):
        self.cub = Path('test_histat.cub')
        isis.hi2isis(img, to=self.cub)

    def tearDown(self):
        self.cub.unlink()
        with contextlib.suppress(FileNotFoundError):
            Path('print.prt').unlink()

    def test_histat_with_to(self):
        tofile = self.cub.with_suffix('.histat')
        isis.histat(self.cub, to=tofile)
        self.assertTrue(tofile.is_file())
        tofile.unlink()

    def test_histat_without_to(self):
        s = isis.histat(self.cub).stdout
        self.assertTrue(s.startswith('Group = IMAGE_POSTRAMP'))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = []

setup_requirements = []

test_requirements = []

setup(
    author="Ross A. Beyer",
    author_email='rbeyer@seti.org',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    description="A Python library to wrap functions and functionality for the Integrated Software for Imagers and Spectrometers (ISIS).",
    install_requires=requirements,
    license="BSD-3-Clause License",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='kalasiris',
    name='kalasiris',
    packages=find_packages(include=['kalasiris']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/rbeyer/kalasiris',
    version='1.7.0',
    zip_safe=False,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Many ISIS programs require the creation of a fromlist file for input.
These functions and classes provide convenient mechanisms for
creating those files, or simply creating temporary versions.
"""

# Copyright 2019-2020, Ross A. Beyer (rbeyer@seti.org)
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import builtins
import os
import tempfile
import sys
import warnings
from pathlib import Path

# with fromlist.open([file1, file2, file3]) as fl:
#     isis.cubeit(fl, to='foo.cub')
#
# with fromlist.open([file1, file2, file3], pathlike='foo') as fl:
#     isis.cubeit(fl, to='foo.cub')
#
# fromlist_path = fromlist.make([file1, file2, file3])
# fromlist_path = fromlist.make([file1, file2, file3], pathlike='foo')


def print(fromlist: list, file=sys.stdout):
    """Works like :func:`print`, but when given a list, will write out
    that list, one element per line.

    This is the format that many ISIS programs which take a
    ``FROMLIST=`` parameter need.

    Therefore, if you wanted to create a fromlist file, you
    could do::

         with open('fromlist.txt', 'w') as f:
             fromlist.print(['a.cub', 'b.cub', 'c.cub'], file=f)

         isis.cubeit(fromlist='fromlist.txt', to='stacked.cub')

    However, it is more likely that you would use :func:`.fromlist.make()`
    or the :class:`.fromlist.temp()` context manager.
    """
    for elem in fromlist:
        builtins.print(str(elem), file=file)


def print_fl(fromlist: list, file=sys.stdout):
    """Synonym for :func:`.fromlist.print()`

    *This function is deprecated, and may be removed
    at the next major patch.*
    """
    warnings.warn('Original syntax, may be removed at next major patch. '
                  'Use fromlist.print() instead.', DeprecationWarning)
    print(fromlist, file)


def make(fromlist: list, pathlike=None) -> Path:
    """Creates a file with the fromlist elements one per line.

    If *pathlike* is given, that will be the path used, otherwise
    create a temporary file and return its path.

    You're responsible for deleting it after you're done.

    You can use it like this::

        fromlist_path = fromlist.make(['a.cub', 'b.cub', 'c.cub'])
        isis.cubeit(fromlist=fromlist_path, to='stacked.cub')
        fromlist_path.unlink()

    However, using the :class:`.fromlist.temp()` context manager might
    be even more handy.
    """
    filelike = None
    mode = 'wt'
    if pathlike is None:
        filelike = tempfile.NamedTemporaryFile(mode=mode, delete=False)
    else:
        filelike = open(pathlike, mode=mode)

    print(fromlist, file=filelike)
    filelike.close()

    return Path(filelike.name)


class open_fl():
    """This is a context manager that works similarly to :func:`open`, but
    for creating fromlist files.  Use it like this::

     with fromlist.open_fl(['a.cub', 'b.cub', 'c.cub']) as f:
         isis.cubeit(fromlist=f.name, to='stacked.cub')

    Its probably better to use :class:`.fromlist.temp()`, however.

    *This context manager is deprecated, and may be removed
    at the next major patch.*
    """
    warnings.warn("Original syntax, may be removed at next major patch. "
                  "Use fromlist.temp() instead.", DeprecationWarning)

    def __init__(self, fromlist: list, pathlike=None):
        self.fromlist = fromlist
        self.name = str(pathlike)
        self.temp = False
        self.file = None

        if pathlike is None:
            self.temp = True

    def __enter__(self):
        p = make(self.fromlist, self.name)
        self.file = open(p, mode='r')
        return self.file

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.file:
            self.file.close()
        if self.temp:
            os.unlink(self.file.name)


class temp():
    """This is a context manager that creates a temporary
    fromlist file and then gets rid of it for you.
    Use it like this::

     with fromlist.temp(['a.cub', 'b.cub', 'c.cub']) as f:
         isis.cubeit(fromlist=f, to='stacked.cub')

    The object that is bound to the *as* clause of the with
    statement is a :class:`pathlib.Path()`.
    """

    def __init__(self, fromlist: list):
        self.path = make(fromlist)

    def __enter__(self) -> Path:
        return self.path

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.path.unlink()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""The ISIS ``cubenorm`` program outputs plain text table information and
also reads it in, but the format is a very specific fixed-width
table format.  A plain :func:`csv.reader` or :func:`csv.DictReader` using
the :class:`.cubenormfile.Dialect` object will be able to read the text
output of ``cubenorm``, but to write out a file that ``cubenorm`` will
read in, you will need to use the :class:`.cubenormfile.writer` or
:class:`.cubenormfile.DictWriter` classes.
"""

# Copyright 2019-2020, Ross A. Beyer (rbeyer@seti.org)
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import csv

# Establish the allowable cubenorm fieldnames and their character widths.
fieldnames = ('Band', 'RowCol', 'ValidPoints', 'Average',
              'Median', 'StdDev', 'Minimum', 'Maximum')

# These widths are extremely fragile, and if cubenorm changes,
# then these will need to be changed, too.
fieldwidth = dict()
for n in fieldnames:
    fieldwidth.setdefault(n, 15)
fieldwidth['Band'] = 8
fieldwidth['RowCol'] = 8


class Dialect(csv.Dialect):
    """A :class:`csv.Dialect` for the output of the ISIS
    ``cubenorm`` program."""
    delimiter = ' '
    skipinitialspace = True
    quoting = csv.QUOTE_NONE
    escapechar = ''
    lineterminator = '\n'


class writer:
    """A class for writing out the fixed-width format required by ``cubenorm``.

    The interface is similar to the :class:`csv.writer` class, but does not
    inheirit from it."""

    def __init__(self, f):
        self.file_object = f

    def writerow(self, row):
        line = ''
        for name, elem in zip(fieldnames, row):
            right_aligned = '{:>' + str(fieldwidth[name]) + '}'
            line += right_aligned.format(elem)

        self.file_object.write(line + '\n')

    def writerows(self, rows):
        for r in rows:
            self.writerow(r)

    def writeheader(self):
        """A convenence function, since the fieldnames are pre-defined."""
        self.writerow(fieldnames)


class DictWriter(csv.DictWriter):
    """A DictWriter for ``cubenorm`` files."""

    def __init__(self, f, restval="", extrasaction="raise",
                 dialect=Dialect, *args, **kwds):
        self.fieldnames = fieldnames
        self.restval = restval
        if extrasaction.lower() not in ("raise", "ignore"):
            raise ValueError("extrasaction (%s) must be 'raise' or 'ignore'"
                             % extrasaction)
        self.extrasaction = extrasaction
        self.writer = writer(f)
